@@ -31,13 +31,13 @@ import threading
 import socket
 import json
 import sys
-import asyncio
 from typing import NamedTuple, Optional, Sequence, List, Dict, Tuple, TYPE_CHECKING, Iterable, Set, Any, TypeVar
 import traceback
 import concurrent
 from concurrent import futures
 import copy
 import functools
+from enum import IntEnum
 
 import aiorpcx
 from aiorpcx import ignore_after
@@ -47,7 +47,7 @@ from . import util
 from .util import (log_exceptions, ignore_exceptions, OldTaskGroup,
                    bfh, make_aiohttp_session, send_exception_to_crash_reporter,
                    is_hash256_str, is_non_negative_integer, MyEncoder, NetworkRetryManager,
-                   nullcontext)
+                   nullcontext, error_text_str_to_safe_str)
 from .bitcoin import COIN
 from . import constants
 from . import blockchain
@@ -83,6 +83,12 @@ NUM_RECENT_SERVERS = 20
 T = TypeVar('T')
 
 
+class ConnectionState(IntEnum):
+    DISCONNECTED  = 0
+    CONNECTING    = 1
+    CONNECTED     = 2
+
+
 def parse_servers(result: Sequence[Tuple[str, str, List[str]]]) -> Dict[str, dict]:
     """Convert servers list (from protocol method "server.peers.subscribe") into dict format.
     Also validate values, such as IP addresses and ports.
@@ -236,8 +242,9 @@ class UntrustedServerReturnedError(NetworkException):
         return _("The server returned an error.")
 
     def __repr__(self):
+        e = self.original_exception
         return (f"<UntrustedServerReturnedError "
-                f"[DO NOT TRUST THIS MESSAGE] original_exception: {repr(self.original_exception)}>")
+                f"[DO NOT TRUST THIS MESSAGE] original_exception: {error_text_str_to_safe_str(repr(e))}>")
 
 
 _INSTANCE = None
@@ -295,18 +302,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
         self._allowed_protocols = {PREFERRED_NETWORK_PROTOCOL}
 
-        # Server for addresses and transactions
-        self.default_server = self.config.get('server', None)
-        # Sanitize default server
-        if self.default_server:
-            try:
-                self.default_server = ServerAddr.from_str(self.default_server)
-            except:
-                self.logger.warning('failed to parse server-string; falling back to localhost:1:s.')
-                self.default_server = ServerAddr.from_str("localhost:1:s")
-        else:
-            self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
-        assert isinstance(self.default_server, ServerAddr), f"invalid type for default_server: {self.default_server!r}"
+        self._init_parameters_from_config()
 
         self.taskgroup = None
 
@@ -338,21 +334,17 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self.interfaces = {}  # these are the ifaces in "initialised and usable" state
         self._closing_ifaces = set()
 
-        self.auto_connect = self.config.get('auto_connect', True)
-        self.proxy = None
-        self._maybe_set_oneserver()
-
         # Dump network messages (all interfaces).  Set at runtime from the console.
         self.debug = False
 
-        self._set_status('disconnected')
+        self._set_status(ConnectionState.DISCONNECTED)
         self._has_ever_managed_to_connect_to_server = False
+        self._was_started = False
 
         # lightning network
         if self.config.get('run_watchtower', False):
             from . import lnwatcher
             self.local_watchtower = lnwatcher.WatchTower(self)
-            self.local_watchtower.adb.start_network(self)
             asyncio.ensure_future(self.local_watchtower.start_watching())
 
     def has_internet_connection(self) -> bool:
@@ -411,7 +403,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 data = f.read()
                 servers_list = json.loads(data)
             return [ServerAddr.from_str(s) for s in servers_list]
-        except:
+        except Exception:
             return []
 
     @with_recent_servers_lock
@@ -423,7 +415,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         try:
             with open(path, "w", encoding='utf-8') as f:
                 f.write(s)
-        except:
+        except Exception:
             pass
 
     async def _server_is_lagging(self) -> bool:
@@ -439,14 +431,22 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     def _set_status(self, status):
         self.connection_status = status
-        self.notify('status')
+        util.trigger_callback('status')
 
     def is_connected(self):
         interface = self.interface
         return interface is not None and interface.is_connected_and_ready()
 
     def is_connecting(self):
-        return self.connection_status == 'connecting'
+        return self.connection_status == ConnectionState.CONNECTING
+
+    def get_connection_status_for_GUI(self):
+        ConnectionStates = {
+            ConnectionState.DISCONNECTED: _('Disconnected'),
+            ConnectionState.CONNECTING: _('Connecting'),
+            ConnectionState.CONNECTED: _('Connected'),
+        }
+        return ConnectionStates[self.connection_status]
 
     async def _request_server_info(self, interface: 'Interface'):
         await interface.ready
@@ -454,7 +454,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
         async def get_banner():
             self.banner = await interface.get_server_banner()
-            self.notify('banner')
+            util.trigger_callback('banner', self.banner)
         async def get_donation_address():
             self.donation_address = await interface.get_donation_address()
         async def get_server_peers():
@@ -464,7 +464,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             server_peers = server_peers[:max_accepted_peers]
             # note that 'parse_servers' also validates the data (which is untrusted input!)
             self.server_peers = parse_servers(server_peers)
-            self.notify('servers')
+            util.trigger_callback('servers', self.get_servers())
         async def get_relay_fee():
             self.relay_fee = await interface.get_relay_fee()
 
@@ -480,28 +480,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         histogram = await interface.get_fee_histogram()
         self.config.mempool_fees = histogram
         self.logger.info(f'fee_histogram {histogram}')
-        self.notify('fee_histogram')
-
-    def get_status_value(self, key):
-        if key == 'status':
-            value = self.connection_status
-        elif key == 'banner':
-            value = self.banner
-        elif key == 'fee':
-            value = self.config.fee_estimates
-        elif key == 'fee_histogram':
-            value = self.config.mempool_fees
-        elif key == 'servers':
-            value = self.get_servers()
-        else:
-            raise Exception('unexpected trigger key {}'.format(key))
-        return value
-
-    def notify(self, key):
-        if key in ['status', 'updated']:
-            util.trigger_callback(key)
-        else:
-            util.trigger_callback(key, self.get_status_value(key))
+        util.trigger_callback('fee_histogram', self.config.mempool_fees)
 
     def get_parameters(self) -> NetworkParameters:
         return NetworkParameters(server=self.default_server,
@@ -509,6 +488,12 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                                  auto_connect=self.auto_connect,
                                  oneserver=self.oneserver)
 
+    def _init_parameters_from_config(self) -> None:
+        self.auto_connect = self.config.get('auto_connect', True)
+        self._set_default_server()
+        self._set_proxy(deserialize_proxy(self.config.get('proxy')))
+        self._maybe_set_oneserver()
+
     def get_donation_address(self):
         if self.is_connected():
             return self.donation_address
@@ -518,6 +503,10 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         with self.interfaces_lock:
             return list(self.interfaces)
 
+    def get_status(self):
+        n = len(self.get_interfaces())
+        return _("Connected to {0} nodes.").format(n) if n > 1 else _("Connected to {0} node.").format(n) if n == 1 else _("Not connected")
+
     def get_fee_estimates(self):
         from statistics import median
         from .simple_config import FEE_ETA_TARGETS
@@ -527,7 +516,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 for n in FEE_ETA_TARGETS:
                     try:
                         out[n] = int(median(filter(None, [i.fee_estimates_eta.get(n) for i in self.interfaces.values()])))
-                    except:
+                    except Exception:
                         continue
                 return out
         else:
@@ -543,7 +532,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         if not hasattr(self, "_prev_fee_est") or self._prev_fee_est != fee_est:
             self._prev_fee_est = copy.copy(fee_est)
             self.logger.info(f'fee_estimates {fee_est}')
-        self.notify('fee')
+        util.trigger_callback('fee', self.config.fee_estimates)
 
     @with_recent_servers_lock
     def get_servers(self):
@@ -599,11 +588,33 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             return server
         return None
 
+    def _set_default_server(self) -> None:
+        # Server for addresses and transactions
+        server = self.config.get('server', None)
+        # Sanitize default server
+        if server:
+            try:
+                self.default_server = ServerAddr.from_str(server)
+            except Exception:
+                self.logger.warning(f'failed to parse server-string ({server!r}); falling back to localhost:1:s.')
+                self.default_server = ServerAddr.from_str("localhost:1:s")
+        else:
+            self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
+        assert isinstance(self.default_server, ServerAddr), f"invalid type for default_server: {self.default_server!r}"
+
     def _set_proxy(self, proxy: Optional[dict]):
         self.proxy = proxy
         dns_hacks.configure_dns_depending_on_proxy(bool(proxy))
         self.logger.info(f'setting proxy {proxy}')
-        util.trigger_callback('proxy_set', self.proxy)
+
+        self.tor_proxy = False
+        if bool(proxy) and proxy['mode'] == 'socks5':
+            # test for Tor
+            self.tor_proxy = util.is_tor_socks_port(proxy['host'], int(proxy['port']))
+            if self.tor_proxy:
+                self.logger.info(f'Proxy is TOR')
+
+        util.trigger_callback('proxy_set', self.proxy, self.tor_proxy)
 
     @log_exceptions
     async def set_parameters(self, net_params: NetworkParameters):
@@ -615,7 +626,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             if proxy:
                 proxy_modes.index(proxy['mode']) + 1
                 int(proxy['port'])
-        except:
+        except Exception:
             return
         self.config.set_key('auto_connect', net_params.auto_connect, False)
         self.config.set_key('oneserver', net_params.oneserver, False)
@@ -627,14 +638,19 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 or self.config.get('oneserver') != net_params.oneserver:
             return
 
+        proxy_changed = self.proxy != proxy
+        oneserver_changed = self.oneserver != net_params.oneserver
+        default_server_changed = self.default_server != server
+        self._init_parameters_from_config()
+        if not self._was_started:
+            return
+
         async with self.restart_lock:
-            self.auto_connect = net_params.auto_connect
-            if self.proxy != proxy or self.oneserver != net_params.oneserver:
-                # Restart the network defaulting to the given server
+            if proxy_changed or oneserver_changed:
+                # Restart the network
                 await self.stop(full_shutdown=False)
-                self.default_server = server
                 await self._start()
-            elif self.default_server != server:
+            elif default_server_changed:
                 await self.switch_to_interface(server)
             else:
                 await self.switch_lagging_interface()
@@ -729,7 +745,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             util.trigger_callback('default_server_changed')
             self.default_server_changed_event.set()
             self.default_server_changed_event.clear()
-            self._set_status('connected')
+            self._set_status(ConnectionState.CONNECTED)
             util.trigger_callback('network_updated')
             if blockchain_updated:
                 util.trigger_callback('blockchain_updated')
@@ -767,7 +783,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         We distinguish by whether it is in self.interfaces.'''
         if not interface: return
         if interface.server == self.default_server:
-            self._set_status('disconnected')
+            self._set_status(ConnectionState.DISCONNECTED)
         await self._close_interface(interface)
         util.trigger_callback('network_updated')
 
@@ -790,7 +806,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self._connecting_ifaces.add(server)
         if server == self.default_server:
             self.logger.info(f"connecting to {server} as new interface")
-            self._set_status('connecting')
+            self._set_status(ConnectionState.CONNECTING)
         self._trying_addr_now(server)
 
         interface = Interface(network=self, server=server, proxy=self.proxy)
@@ -886,10 +902,14 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     @best_effort_reliable
     @catch_server_exceptions
     async def get_merkle_for_transaction(self, tx_hash: str, tx_height: int) -> dict:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_merkle_for_transaction(tx_hash=tx_hash, tx_height=tx_height)
 
     @best_effort_reliable
     async def broadcast_transaction(self, tx: 'Transaction', *, timeout=None) -> None:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         if timeout is None:
             timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
         try:
@@ -898,14 +918,15 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         except (RequestTimedOut, asyncio.CancelledError, asyncio.TimeoutError):
             raise  # pass-through
         except aiorpcx.jsonrpc.CodeMessageError as e:
-            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {repr(e)}")
+            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
             raise TxBroadcastServerReturnedError(self.sanitize_tx_broadcast_response(e.message)) from e
         except BaseException as e:  # intentional BaseException for sanity!
-            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {repr(e)}")
+            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
             send_exception_to_crash_reporter(e)
             raise TxBroadcastUnknownError() from e
         if out != tx.txid():
-            self.logger.info(f"unexpected txid for broadcast_transaction [DO NOT TRUST THIS MESSAGE]: {out} != {tx.txid()}")
+            self.logger.info(f"unexpected txid for broadcast_transaction [DO NOT TRUST THIS MESSAGE]: "
+                             f"{error_text_str_to_safe_str(out)} != {tx.txid()}")
             raise TxBroadcastHashMismatch(_("Server returned unexpected transaction ID."))
 
     async def try_broadcasting(self, tx, name) -> bool:
@@ -1098,31 +1119,43 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     @best_effort_reliable
     @catch_server_exceptions
     async def request_chunk(self, height: int, tip=None, *, can_return_early=False):
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.request_chunk(height, tip=tip, can_return_early=can_return_early)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_transaction(self, tx_hash: str, *, timeout=None) -> str:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_transaction(tx_hash=tx_hash, timeout=timeout)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_history_for_scripthash(self, sh: str) -> List[dict]:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_history_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def listunspent_for_scripthash(self, sh: str) -> List[dict]:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.listunspent_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_balance_for_scripthash(self, sh: str) -> dict:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_balance_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_txid_from_txpos(self, tx_height, tx_pos, merkle):
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_txid_from_txpos(tx_height, tx_pos, merkle)
 
     def blockchain(self) -> Blockchain:
@@ -1185,7 +1218,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         interface = self.interface
         return interface.tip if interface else 0
 
-    def get_local_height(self):
+    def get_local_height(self) -> int:
         """Length of header chain, POW-verified.
         In case of a chain split, this is for the branch the main interface is on,
         but it is the tip of that branch (even if main interface is behind).
@@ -1208,8 +1241,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         assert not self._closing_ifaces
         self.logger.info('starting network')
         self._clear_addr_retry_times()
-        self._set_proxy(deserialize_proxy(self.config.get('proxy')))
-        self._maybe_set_oneserver()
+        self._init_parameters_from_config()
         await self.taskgroup.spawn(self._run_new_interface(self.default_server))
 
         async def main():
@@ -1234,11 +1266,15 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         Note: the jobs will *restart* every time the network restarts, e.g. on proxy
         setting changes.
         """
+        self._was_started = True
         self._jobs = jobs or []
         asyncio.run_coroutine_threadsafe(self._start(), self.asyncio_loop)
 
     @log_exceptions
     async def stop(self, *, full_shutdown: bool = True):
+        if not self._was_started:
+            self.logger.info("not stopping network as it was never started")
+            return
         self.logger.info("stopping network")
         # timeout: if full_shutdown, it is up to the caller to time us out,
         #          otherwise if e.g. restarting due to proxy changes, we time out fast
@@ -1326,7 +1362,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                     async with session.post(url, json=json, headers=headers) as resp:
                         return await on_finish(resp)
             else:
-                assert False
+                raise Exception(f"unexpected {method=!r}")
 
     @classmethod
     def send_http_on_proxy(cls, method, url, **kwargs):
