@@ -46,6 +46,7 @@ from ipaddress import IPv4Address, IPv6Address
 import random
 import secrets
 import functools
+from functools import partial
 from abc import abstractmethod, ABC
 import socket
 
@@ -103,6 +104,7 @@ def decimal_point_to_base_unit_name(dp: int) -> str:
 
 
 def base_unit_name_to_decimal_point(unit_name: str) -> int:
+    """Returns the max number of digits allowed after the decimal point."""
     # e.g. "BTC" -> 8
     try:
         return base_units[unit_name]
@@ -185,7 +187,10 @@ class FileExportFailed(Exception):
         return _("Failed to export to file.") + "\n" + self.message
 
 
-class WalletFileException(Exception): pass
+class WalletFileException(Exception):
+    def __init__(self, message='', *, should_report_crash: bool = False):
+        Exception.__init__(self, message)
+        self.should_report_crash = should_report_crash
 
 
 class BitcoinException(Exception): pass
@@ -432,7 +437,7 @@ def json_encode(obj):
 def json_decode(x):
     try:
         return json.loads(x, parse_float=Decimal)
-    except:
+    except Exception:
         return x
 
 def json_normalize(x):
@@ -449,15 +454,22 @@ def constant_time_compare(val1, val2):
     return hmac.compare_digest(to_bytes(val1, 'utf8'), to_bytes(val2, 'utf8'))
 
 
-# decorator that prints execution time
 _profiler_logger = _logger.getChild('profiler')
-def profiler(func):
+def profiler(func=None, *, min_threshold: Union[int, float, None] = None):
+    """Function decorator that logs execution time.
+
+    min_threshold: if set, only log if time taken is higher than threshold
+    NOTE: does not work with async methods.
+    """
+    if func is None:
+        return partial(profiler, min_threshold=min_threshold)
     def do_profile(args, kw_args):
         name = func.__qualname__
         t0 = time.time()
         o = func(*args, **kw_args)
         t = time.time() - t0
-        _profiler_logger.debug(f"{name} {t:,.4f} sec")
+        if min_threshold is None or t > min_threshold:
+            _profiler_logger.debug(f"{name} {t:,.4f} sec")
         return o
     return lambda *args, **kw_args: do_profile(args, kw_args)
 
@@ -562,7 +574,7 @@ def assert_bytes(*args):
     try:
         for x in args:
             assert isinstance(x, (bytes, bytearray))
-    except:
+    except Exception:
         print('assert bytes failed', list(map(type, args)))
         raise
 
@@ -646,7 +658,7 @@ def is_hex_str(text: Any) -> bool:
     if not isinstance(text, str): return False
     try:
         b = bytes.fromhex(text)
-    except:
+    except Exception:
         return False
     # forbid whitespaces in text:
     if len(text) != 2 * len(b):
@@ -1191,7 +1203,7 @@ def parse_json(message):
         return None, message
     try:
         j = json.loads(message[0:n].decode('utf8'))
-    except:
+    except Exception:
         j = None
     return j, message[n+1:]
 
@@ -1340,6 +1352,15 @@ class TxMinedInfo(NamedTuple):
             return f"{self.height}x{self.txpos}"
         return None
 
+    def is_local_like(self) -> bool:
+        """Returns whether the tx is local-like (LOCAL/FUTURE)."""
+        from .address_synchronizer import TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_UNCONF_PARENT
+        if self.height > 0:
+            return False
+        if self.height in (TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_UNCONF_PARENT):
+            return False
+        return True
+
 
 class ShortID(bytes):
 
@@ -1548,8 +1569,20 @@ class NetworkJobOnDefaultServer(Logger, ABC):
         self.reset_request_counters()
 
     async def _start(self, interface: 'Interface'):
+        self.logger.debug(f"starting. interface.server={repr(str(interface.server))}")
         self.interface = interface
-        await interface.taskgroup.spawn(self._run_tasks(taskgroup=self.taskgroup))
+
+        taskgroup = self.taskgroup
+        async def run_tasks_wrapper():
+            self.logger.debug(f"starting taskgroup ({hex(id(taskgroup))}).")
+            try:
+                await self._run_tasks(taskgroup=taskgroup)
+            except Exception as e:
+                self.logger.error(f"taskgroup died ({hex(id(taskgroup))}). exc={e!r}")
+                raise
+            finally:
+                self.logger.debug(f"taskgroup stopped ({hex(id(taskgroup))}).")
+        await interface.taskgroup.spawn(run_tasks_wrapper)
 
     @abstractmethod
     async def _run_tasks(self, *, taskgroup: OldTaskGroup) -> None:
@@ -1562,6 +1595,7 @@ class NetworkJobOnDefaultServer(Logger, ABC):
             raise asyncio.CancelledError()
 
     async def stop(self, *, full_shutdown: bool = True):
+        self.logger.debug(f"stopping. {full_shutdown=}")
         if full_shutdown:
             unregister_callback(self._restart)
         await self.taskgroup.cancel_remaining()
